@@ -5,6 +5,7 @@ from __future__ import division, unicode_literals, print_function, absolute_impo
 import copy
 import math
 import operator as op
+import warnings
 
 from pint import DimensionalityError, OffsetUnitCalculusError, UnitRegistry
 from pint.unit import UnitsContainer
@@ -95,18 +96,21 @@ class TestQuantity(QuantityTestCase):
                              ('{0.magnitude!s}',  str(x.magnitude)), ('{0.units!s}',  str(x.units)),
                              ('{0.magnitude!r}',  repr(x.magnitude)), ('{0.units!r}',  repr(x.units)),
                              ('{0:.4f}', '{0:.4f} {1!s}'.format(x.magnitude, x.units)),
-                             ('{0:L}', r'4.12345678 \frac{kilogram \cdot meter^{2}}{second}'),
+                             ('{0:L}', r'4.12345678\ \frac{\mathrm{kilogram} \cdot \mathrm{meter}^{2}}{\mathrm{second}}'),
                              ('{0:P}', '4.12345678 kilogram·meter²/second'),
                              ('{0:H}', '4.12345678 kilogram meter<sup>2</sup>/second'),
                              ('{0:C}', '4.12345678 kilogram*meter**2/second'),
                              ('{0:~}', '4.12345678 kg * m ** 2 / s'),
-                             ('{0:L~}', r'4.12345678 \frac{kg \cdot m^{2}}{s}'),
+                             ('{0:L~}', r'4.12345678\ \frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}'),
                              ('{0:P~}', '4.12345678 kg·m²/s'),
                              ('{0:H~}', '4.12345678 kg m<sup>2</sup>/s'),
                              ('{0:C~}', '4.12345678 kg*m**2/s'),
                              ('{0:Lx}', r'\SI[]{4.12345678}{\kilo\gram\meter\squared\per\second}'),
                              ):
             self.assertEqual(spec.format(x), result)
+        # Check the special case that prevents e.g. '3 1 / second'
+        x = self.Q_(3, UnitsContainer(second=-1))
+        self.assertEqual('{0}'.format(x), '3 / second')
 
     def test_format_compact(self):
         q1 = (200e-9 * self.ureg.s).to_compact()
@@ -119,19 +123,25 @@ class TestQuantity(QuantityTestCase):
         self.assertEqual(q2.magnitude, q2b.magnitude)
         self.assertEqual(q2.units, q2b.units)
 
+        q3 = (-1000.0 * self.ureg('meters')).to_compact()
+        q3b = self.Q_(-1., 'kilometer')
+        self.assertEqual(q3.magnitude, q3b.magnitude)
+        self.assertEqual(q3.units, q3b.units)
+
         self.assertEqual('{0:#.1f}'.format(q1), '{0}'.format(q1b))
         self.assertEqual('{0:#.1f}'.format(q2), '{0}'.format(q2b))
+        self.assertEqual('{0:#.1f}'.format(q3), '{0}'.format(q3b))
 
 
     def test_default_formatting(self):
         ureg = UnitRegistry()
         x = ureg.Quantity(4.12345678, UnitsContainer(meter=2, kilogram=1, second=-1))
-        for spec, result in (('L', r'4.12345678 \frac{kilogram \cdot meter^{2}}{second}'),
+        for spec, result in (('L', r'4.12345678\ \frac{\mathrm{kilogram} \cdot \mathrm{meter}^{2}}{\mathrm{second}}'),
                              ('P', '4.12345678 kilogram·meter²/second'),
                              ('H', '4.12345678 kilogram meter<sup>2</sup>/second'),
                              ('C', '4.12345678 kilogram*meter**2/second'),
                              ('~', '4.12345678 kg * m ** 2 / s'),
-                             ('L~', r'4.12345678 \frac{kg \cdot m^{2}}{s}'),
+                             ('L~', r'4.12345678\ \frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}'),
                              ('P~', '4.12345678 kg·m²/s'),
                              ('H~', '4.12345678 kg m<sup>2</sup>/s'),
                              ('C~', '4.12345678 kg*m**2/s'),
@@ -178,6 +188,17 @@ class TestQuantity(QuantityTestCase):
             self.assertIsNot(r, q)
             self.assertIsNot(r._magnitude, a)
 
+    @helpers.requires_numpy()
+    def test_retain_unit(self):
+        # Test that methods correctly retain units and do not degrade into
+        # ordinary ndarrays.  List contained in __copy_units.
+        a = np.ones((3, 2))
+        q = self.Q_(a, "km")
+        self.assertEqual(q.u, q.reshape(2, 3).u)
+        self.assertEqual(q.u, q.swapaxes(0, 1).u)
+        self.assertEqual(q.u, q.mean().u)
+        self.assertEqual(q.u, np.compress((q==q[0,0]).any(0), q).u)
+
     def test_context_attr(self):
         self.assertEqual(self.ureg.meter, self.Q_(1, 'meter'))
 
@@ -195,6 +216,11 @@ class TestQuantity(QuantityTestCase):
         self.assertEqual(self.Q_(1, 'meter')/self.Q_(1, 'meter'), 1)
         self.assertEqual((self.Q_(1, 'meter')/self.Q_(1, 'mm')).to(''), 1000)
 
+        self.assertEqual(self.Q_(10) // self.Q_(360, 'degree'), 1)
+        self.assertEqual(self.Q_(400, 'degree') // self.Q_(2 * math.pi), 1)
+        self.assertEqual(self.Q_(400, 'degree') // (2 * math.pi), 1)
+        self.assertEqual(7 // self.Q_(360, 'degree'), 1)
+
     def test_offset(self):
         self.assertQuantityAlmostEqual(self.Q_(0, 'kelvin').to('kelvin'), self.Q_(0, 'kelvin'))
         self.assertQuantityAlmostEqual(self.Q_(0, 'degC').to('kelvin'), self.Q_(273.15, 'kelvin'))
@@ -297,6 +323,15 @@ class TestQuantityToCompact(QuantityTestCase):
         self.compareQuantity_compact(self.Q_(101.3e3, 'kg/m/s^2'),
             101.3*ureg.kPa, ureg.Pa)
 
+    def test_limits_magnitudes(self):
+        ureg = self.ureg
+        self.compareQuantity_compact(0*ureg.m, 0*ureg.m)
+        self.compareQuantity_compact(float('inf')*ureg.m, float('inf')*ureg.m)
+
+    def test_nonnumeric_magnitudes(self):
+        ureg = self.ureg
+        x = "some string"*ureg.m
+        self.assertRaises(RuntimeError, self.compareQuantity_compact(x,x))
 
 class TestQuantityBasicMath(QuantityTestCase):
 
@@ -414,14 +449,76 @@ class TestQuantityBasicMath(QuantityTestCase):
         func(op.itruediv, '4.2*meter', '10*inch', '0.42*meter/inch', unit)
 
     def _test_quantity_floordiv(self, unit, func):
-        func(op.floordiv, unit * 10.0, '4.2*meter', '2/meter', unit)
-        func(op.floordiv, '24*meter', unit * 10.0, '2*meter', unit)
-        func(op.floordiv, '10*meter', '4.2*inch', '2*meter/inch', unit)
+        a = self.Q_('10*meter')
+        b = self.Q_('3*second')
+        self.assertRaises(DimensionalityError, op.floordiv, a, b)
+        self.assertRaises(DimensionalityError, op.floordiv, 3, b)
+        self.assertRaises(DimensionalityError, op.floordiv, a, 3)
+        self.assertRaises(DimensionalityError, op.ifloordiv, a, b)
+        self.assertRaises(DimensionalityError, op.ifloordiv, 3, b)
+        self.assertRaises(DimensionalityError, op.ifloordiv, a, 3)
+        func(op.floordiv, unit * 10.0, '4.2*meter/meter', 2, unit)
+        func(op.floordiv, '10*meter', '4.2*inch', 93, unit)
+
+    def _test_quantity_mod(self, unit, func):
+        a = self.Q_('10*meter')
+        b = self.Q_('3*second')
+        self.assertRaises(DimensionalityError, op.mod, a, b)
+        self.assertRaises(DimensionalityError, op.mod, 3, b)
+        self.assertRaises(DimensionalityError, op.mod, a, 3)
+        self.assertRaises(DimensionalityError, op.imod, a, b)
+        self.assertRaises(DimensionalityError, op.imod, 3, b)
+        self.assertRaises(DimensionalityError, op.imod, a, 3)
+        func(op.mod, unit * 10.0, '4.2*meter/meter', 1.6, unit)
 
     def _test_quantity_ifloordiv(self, unit, func):
-        func(op.ifloordiv, 10.0, '4.2*meter', '2/meter', unit)
-        func(op.ifloordiv, '24*meter', 10.0, '2*meter', unit)
-        func(op.ifloordiv, '10*meter', '4.2*inch', '2*meter/inch', unit)
+        func(op.ifloordiv, 10.0, '4.2*meter/meter', 2, unit)
+        func(op.ifloordiv, '10*meter', '4.2*inch', 93, unit)
+
+    def _test_quantity_divmod_one(self, a, b):
+        if isinstance(a, string_types):
+            a = self.Q_(a)
+        if isinstance(b, string_types):
+            b = self.Q_(b)
+
+        q, r = divmod(a, b)
+        self.assertEqual(q, a // b)
+        self.assertEqual(r, a % b)
+        self.assertEqual(a, (q * b) + r)
+        self.assertEqual(q, math.floor(q))
+        if b > (0 * b):
+            self.assertTrue((0 * b) <= r < b)
+        else:
+            self.assertTrue((0 * b) >= r > b)
+        if isinstance(a, self.Q_):
+            self.assertEqual(r.units, a.units)
+        else:
+            self.assertTrue(r.unitless)
+        self.assertTrue(q.unitless)
+
+        copy_a = copy.copy(a)
+        a %= b
+        self.assertEqual(a, r)
+        copy_a //= b
+        self.assertEqual(copy_a, q)
+
+    def _test_quantity_divmod(self):
+        self._test_quantity_divmod_one('10*meter', '4.2*inch')
+        self._test_quantity_divmod_one('-10*meter', '4.2*inch')
+        self._test_quantity_divmod_one('-10*meter', '-4.2*inch')
+        self._test_quantity_divmod_one('10*meter', '-4.2*inch')
+
+        self._test_quantity_divmod_one('400*degree', '3')
+        self._test_quantity_divmod_one('4', '180 degree')
+        self._test_quantity_divmod_one(4, '180 degree')
+        self._test_quantity_divmod_one('20', 4)
+        self._test_quantity_divmod_one('300*degree', '100 degree')
+
+        a = self.Q_('10*meter')
+        b = self.Q_('3*second')
+        self.assertRaises(DimensionalityError, divmod, a, b)
+        self.assertRaises(DimensionalityError, divmod, 3, b)
+        self.assertRaises(DimensionalityError, divmod, a, 3)
 
     def _test_numeric(self, unit, ifunc):
         self._test_quantity_add_sub(unit, self._test_not_inplace)
@@ -429,6 +526,8 @@ class TestQuantityBasicMath(QuantityTestCase):
         self._test_quantity_mul_div(unit, self._test_not_inplace)
         self._test_quantity_imul_idiv(unit, ifunc)
         self._test_quantity_floordiv(unit, self._test_not_inplace)
+        self._test_quantity_mod(unit, self._test_not_inplace)
+        self._test_quantity_divmod()
         #self._test_quantity_ifloordiv(unit, ifunc)
 
     def test_float(self):
